#' Wrapper for Applying Atomic Functions to Arrays.
#'
#' The Apply function is an extension of the mapply function, which instead of taking lists of unidimensional objects as input, takes lists of multidimensional objects as input, which may have different numbers of dimensions and dimension lengths. The user can specify which dimensions of each array (or matrix) the function is to be applied over with the margins option. 
#' @param data A single object (vector, matrix or array) or a list of objects. They must be in the same order as expected by AtomicFun.
#' @param target_dims List of vectors containing the dimensions to be input into AtomicFun for each of the objects in the data. These vectors can contain either integers specifying the dimension position, or characters corresponding to the dimension names. If both margins and target_dims are specified, margins takes priority over target_dims.
#' @param AtomicFun Function to be applied to the arrays.
#' @param ... Additional arguments to be used in the AtomicFun.
#' @param margins List of vectors containing the margins for the input objects to be split by. Or, if there is a single vector of margins specified and a list of objects in data, then the single set of margins is applied over all objects. These vectors can contain either integers specifying the dimension position, or characters corresponding to the dimension names. If both margins and target_dims are specified, margins takes priority over target_dims.
#' @param parallel Logical, should the function be applied in parallel.
#' @param ncores The number of cores to use for parallel computation.
#' @details When using a single object as input, Apply is almost identical to the apply function. For multiple input objects, the output array will have dimensions equal to the dimensions specified in 'margins'.
#' @return Array or matrix or vector resulting from AtomicFun.
#' @references Wickham, H (2011), The Split-Apply-Combine Strategy for Data Analysis, Journal of Statistical Software.
#' @export
#' @examples
#' #Change in the rate of exceedance for two arrays, with different 
#' #dimensions, for some matrix of exceedances.
#' data = list(array(rnorm(2000), c(10,10,20)), array(rnorm(1000), c(10,10,10)), 
#'             array(rnorm(100), c(10, 10)))
#' test_fun <- function(x, y, z) {((sum(x > z) / (length(x))) / 
#'                                (sum(y > z) / (length(y)))) * 100}
#' margins = list(c(1, 2), c(1, 2), c(1,2))
#' test <- Apply(data, margins, AtomicFun = "test_fun")
Apply <- function(data, target_dims = NULL, AtomicFun, ..., margins = NULL, parallel = FALSE, ncores = NULL) {
  # Check data
  if (!is.list(data)) {
    data <- list(data)
  }
  if (any(!sapply(data, is.numeric))) {
    stop("Parameter 'data' must be one or a list of numeric objects.")
  }
  # Check target_dims and margins
  if (is.null(margins) && is.null(target_dims)) {
    stop("One of 'margins' or 'target_dims' must be specified.")
  }
  if (!is.null(margins)) {
    target_dims <- NULL
  }
  if (!is.null(margins)) {
  # Check margins and build target_dims accordingly
    if (!is.list(margins)) {
      margins <- rep(list(margins), length(data))
    }
    if (any(!sapply(margins, 
                    function(x) is.character(x) || is.numeric(x)))) {
      stop("Parameter 'margins' must be one or a list of numeric or ",
           "character vectors.")
    }
    if (any(sapply(margins, length) == 0)) {
      stop("Parameter 'margins' must not contain length-0 vectors.")
    }
    duplicate_dim_specs <- sapply(margins, 
      function(x) {
        length(unique(x)) != length(x)
      })
    if (any(duplicate_dim_specs)) {
      stop("Parameter 'margins' must not contain duplicated dimension ",
           "specifications.")
    }
    target_dims <- vector('list', length(data))
    for (i in 1 : length(data)) {
      if (is.character(unlist(margins[i]))) {
        margins_new <- margins[[i]]
        for (j in 1 : length(margins_new)) {
          margins_new[j] <- which(names(dim(data[[i]])) == margins_new[j])
        }
        margins[[i]] <- margins_new
      }
      target_dims[[i]] <- c(1 : length(dim(data[[i]])))[-c(margins[[i]])]   
    }
  } else {
  # Check target_dims and build margins accordingly
    if (!is.list(target_dims)) {
      target_dims <- rep(list(target_dims), length(data))
    }
    if (any(!sapply(target_dims, 
                    function(x) is.character(x) || is.numeric(x)))) {
      stop("Parameter 'target_dims' must be one or a list of numeric or ",
           "character vectors.")
    }
    if (any(sapply(target_dims, length) == 0)) {
      stop("Parameter 'target_dims' must not contain length-0 vectors.")
    }
    duplicate_dim_specs <- sapply(target_dims, 
      function(x) {
        length(unique(x)) != length(x)
      })
    if (any(duplicate_dim_specs)) {
      stop("Parameter 'target_dims' must not contain duplicated dimension ",
           "specifications.")
    }
    margins <- vector('list', length(data))
    for (i in 1 : length(data)) {
      if (is.character(unlist(target_dims[i]))) {
        margins2 <- target_dims[[i]]
        margins2_new <- c()
        for (j in 1 : length(margins2)) {
          margins2_new[j] <- which(names(dim(data[[i]])) == margins2[j])
        }
        target_dims[[i]] <- margins2_new
      }
      margins[[i]] <- c(1 : length(dim(data[[i]]))[-c(target_dims[[i]])])
    }
  }
  # Reorder dimensions of input data for target dims to be left-most
  for (i in 1 : length(data)) {
    if (is.unsorted(target_dims[[i]]) || 
        (max(target_dims[[i]]) > length(target_dims[[i]]))) {
      targ_dims <- sort(target_dims[[i]])
      marg_dims <- (1 : length(dim(data[[i]])))[- target_dims[[i]]]
      data[[i]] <- .aperm2(data[[i]], c(targ_dims, marg_dims))
      target_dims[[i]] <- 1 : length(targ_dims)
      margins[[i]] <- c(1 : length(dim(data[[i]])))[-c(target_dims[[i]])]   
    }
  }
  # Check AtomicFun
  if (is.character(AtomicFun)) {
    try({AtomicFun <- get(AtomicFun)}, silent = TRUE)
    if (!is.function(AtomicFun)) {
      stop("Could not find the function '", AtomicFun, "'.")
    }
  }
  if (!is.function(AtomicFun)) {
    stop("Parameter 'AtomicFun' must be a function or a character string ",
         "with the name of a function.")
  }
  # Check parallel
  if (!is.logical(parallel)) {
    stop("Parameter 'parallel' must be logical.")
  }
  # Check ncores
  if (parallel) {
    if (is.null(ncores)) {
      ncores <- availableCores() - 1
    }
    if (!is.numeric(ncores)) {
      stop("Parameter 'ncores' must be numeric.")
    }
    ncores <- round(ncores)
    ncores <- min(availableCores() - 1, ncores)
  }

  names <- names(dim(data[[1]]))[margins[[1]]]
  input <- list()

  splatted_f <- splat(AtomicFun)

  if (!is.null(margins)) {
    #Check margins match for input objects
    all_dims <- c(unlist(lapply(1 : length(data), 
                                function(x) sum(dim(data[[x]])[margins[[x]]]))))
    print(all_dims)
    pos_dim <-  min(which(all_dims == max(all_dims)))
    dim_template <- dim(data[[pos_dim]])[margins[[pos_dim]]]
    print(pos_dim)
    print(dim_template)
    for (i in 1 : length(data)) {
      if (identical(dim(data[[i]])[margins[[i]]], dim_template) == FALSE) {
        for (j in 1 : (length(dim(data[[i]])[margins[[i]]]))) {
          print("OK")
          print(c(dim(data[[i]])[margins[[i]]])[j] != dim_template[j])
          if (c(dim(data[[i]])[margins[[i]]])[j] != dim_template[j]) {
            print(class(data[[i]]))
            data[[i]] <- InsertDim(data[[i]], posdim = margins[[i]][j], lendim = dim_template[j])
            data[[i]] <- adrop(data[[i]], drop = (margins[[i]][j] + 1))
            print(class(data[[i]]))
           }
         }
       }
    }
    print(dim(data)[[2]])
    .isolate <- function(data, margin_length, drop = TRUE) {
      eval(dim(environment()$data))
      structure(list(env = environment(), index = margin_length, subs = as.name("[")), 
                class = c("indexed_array"))
    }
    for (i in 1 : length(data)) {
      margin_length <- lapply(dim(data[[i]]), function(x) 1 : x)
      margin_length[-margins[[i]]] <- ""
      margin_length <- expand.grid(margin_length, KEEP.OUT.ATTRS = FALSE, 
                                   stringsAsFactors = FALSE)
      input[[i]] <- .isolate(data[[i]], margin_length)
    }
    dims <- dim(data[[1]])[margins[[1]]]
    i_max <- length(input[[1]])[1] / dims[[1]]
    k <- length(input[[1]]) / i_max
    if (parallel == TRUE) {
      registerDoParallel(ncores)  
    }
    WrapperFun <- llply(1 : i_max, function(i)
      sapply((k * i - (k - 1)) : (k * i), function(x) splatted_f(lapply(input, `[[`, x),...), simplify = FALSE),
      .parallel = parallel)
    if (parallel == TRUE) {
      registerDoSEQ()
    }
    if (is.null(dim(WrapperFun[[1]][[1]]))) {
      WrapperFun <- array(as.numeric(unlist(WrapperFun)), dim=c(c(length((WrapperFun[[1]])[[1]])),
                                                                dim(data[[1]])[margins[[1]]]))
    } else {
      WrapperFun <- array(as.numeric(unlist(WrapperFun)), dim=c(c(dim(WrapperFun[[1]][[1]])),
                                                                dim(data[[1]])[margins[[1]]]))
    }
  } else {
    WrapperFun <- splatted_f(data, ...)
  }
  if (!is.null(dim(WrapperFun))) {
    names(dim(WrapperFun))[(length(dim(WrapperFun)) - length(names) + 1) : length(dim(WrapperFun))] <- c(names)
  }
  out <- WrapperFun
}
